import requests
import time
import sys

from loguru import logger

from KnowAlmost.utils.error import (
    BadRequest, Forbidden, HTTPException, NotFound, TooManyRequests,
    KnowAlmostException, Unauthorized, KnowAlmostServerError
)
from KnowAlmost.utils.util import get_x96
from setting import LIST_PARAMS, LIST_HEADERS
from urllib.parse import urlencode


class KnowAlmost:
    """KnowAlmost API v2.0.1 Interface
    """

    def __init__(
        self, *, host='www.zhihu.com',
        proxy=None, retry_count=0, retry_delay=0, retry_errors=None,
        timeout=60,
        wait_on_rate_limit=False
    ):
        self.host = host
        self.proxy = {}
        if proxy is not None:
            self.proxy['https'] = proxy

        self.retry_count = retry_count
        self.retry_delay = retry_delay
        self.retry_errors = retry_errors
        self.timeout = timeout

        self.wait_on_rate_limit = wait_on_rate_limit

        self.session = requests.Session()

    def request(
        self, endpoint_parameters=(), params=None, query="", token="",
        headers=None, offset=0, search_source="Normal",
        **kwargs
    ):
        if headers is None:
            headers = LIST_HEADERS
        headers["Cookie"] = f"d_c0={token};"

        # Build the request URL
        path = '/api/v4/search_v3'
        url = 'https://' + self.host + path

        if params is None:
            params = LIST_PARAMS
        params['q'] = query
        params['offset'] = offset
        params['search_source'] = search_source

        for k, arg in kwargs.items():
            if arg is None:
                continue
            if k not in endpoint_parameters + (
                "gk_version", "t", "correction", "filter_fields", "lc_idx", "show_all_topics", "limit"
            ):
                logger.warning(f'Unexpected parameter: {k}')
            else:
                params[k] = str(arg)

            if k in (
                    "user-agent", "x-api-version", "x-zse-93"
            ):
                headers[k] = str(arg)
        xzse_96 = get_x96(token=token, search_api=path+'?'+urlencode(params))
        headers["x-zse-96"] = xzse_96
        logger.info(f"PARAMS: {params}")
        logger.info(f"HEADERS: {headers}")

        remaining_calls = None
        reset_time = None

        try:
            # Continue attempting request until successful
            retries_performed = 0
            while retries_performed <= self.retry_count:
                if (self.wait_on_rate_limit and
                        reset_time is not None and remaining_calls is not None and remaining_calls < 1):
                    # Handle running out of API calls
                    sleep_time = reset_time - int(time.time())
                    if sleep_time > 0:
                        logger.warning(f"Rate limit reached. Sleeping for: {sleep_time}")
                        time.sleep(sleep_time + 1)  # Sleep for extra sec

                # Execute request
                try:
                    resp = self.session.request(
                        "GET", url, params=params, headers=headers,
                        timeout=self.timeout, proxies=self.proxy
                    )

                except Exception as e:
                    raise KnowAlmostException(f'Failed to send request: {e}').with_traceback(sys.exc_info()[2])

                if 200 <= resp.status_code < 300:
                    break

                retry_delay = self.retry_delay
                if resp.status_code in (420, 429) and self.wait_on_rate_limit:
                    if remaining_calls == 0:
                        # If ran out of calls before waiting switching retry last call
                        continue
                    if 'retry-after' in resp.headers:
                        retry_delay = float(resp.headers['retry-after'])
                elif self.retry_errors and resp.status_code not in self.retry_errors:
                    # Exit request loop if non-retry error code
                    break

                # Sleep before retrying request again
                time.sleep(retry_delay)
                retries_performed += 1

            # If an error was returned, throw an exception
            self.last_response = resp
            if resp.status_code == 400:
                raise BadRequest(resp)
            if resp.status_code == 401:
                raise Unauthorized(resp)
            if resp.status_code == 403:
                raise Forbidden(resp)
            if resp.status_code == 404:
                raise NotFound(resp)
            if resp.status_code == 429:
                raise TooManyRequests(resp)
            if resp.status_code >= 500:
                raise KnowAlmostServerError(resp)
            if resp.status_code and not 200 <= resp.status_code < 300:
                raise HTTPException(resp)

            # Parse the response payload
            return self.last_response
        finally:
            self.session.close()


know_almost_reader = KnowAlmost()



import json as json_lib

from KnowAlmost.utils.error import KnowAlmostException


class Parser:

    def parse(self, payload, *args, **kwargs):
        """
        Parse the response payload and return the result.
        Returns a tuple that contains the result data and the cursors
        (or None if not present).
        """
        raise NotImplementedError


class JSONParser(Parser):

    payload_format = 'json'

    def parse(self, payload, *, return_cursors=False, **kwargs):
        try:
            json = json_lib.loads(payload)
        except Exception as e:
            raise KnowAlmostException(f'Failed to parse JSON payload: {e}')

        if return_cursors and isinstance(json, dict):
            if 'next' in json:
                return json, json['next']
            elif 'next_cursor' in json:
                if 'previous_cursor' in json:
                    cursors = json['previous_cursor'], json['next_cursor']
                    return json, cursors
                else:
                    return json, json['next_cursor']
        return json


class ModelParser(JSONParser):

    def __init__(self, model_factory=None):
        JSONParser.__init__(self)
        self.model_factory = model_factory or ModelFactory

    def parse(self, payload, *, api=None, payload_list=False,
              payload_type=None, return_cursors=False):
        try:
            if payload_type is None:
                return
            model = getattr(self.model_factory, payload_type)
        except AttributeError:
            raise KnowAlmostException(
                f'No model for this payload type: {payload_type}'
            )

        json = JSONParser.parse(self, payload, return_cursors=return_cursors)
        if isinstance(json, tuple):
            json, cursors = json
        else:
            cursors = None

        try:
            if payload_list:
                result = model.parse_list(api, json)
            else:
                result = model.parse(api, json)
        except KeyError:
            raise KnowAlmostException(
                f"Unable to parse response payload: {json}"
            ) from None

        if cursors:
            return result, cursors
        else:
            return result


# KnowAlmost
# Copyright 2009-2022 Joshua Roesslein
# See LICENSE for details.

import requests


class KnowAlmostException(Exception):
    """Base exception for KnowAlmost

    .. version:: 1.0    """
    pass


class HTTPException(KnowAlmostException):
    """HTTPException()

    Exception raised when an HTTP request fails

    .. version:: 1.0
        ``response`` attribute can be an instance of
        :class:`aiohttp.ClientResponse`

    Attributes
    ----------
    response : requests.Response | aiohttp.ClientResponse
        Requests Response from the KnowAlmost API
    api_errors : list[dict[str, int | str]]
        The errors the KnowAlmost API responded with, if any
    api_codes : list[int]
        The error codes the KnowAlmost API responded with, if any
    api_messages : list[str]
        The error messages the KnowAlmost API responded with, if any
    """

    def __init__(self, response, *, response_json=None):
        self.response = response

        self.api_errors = []
        self.api_codes = []
        self.api_messages = []

        try:
            status_code = response.status_code
        except AttributeError:
            # response is an instance of aiohttp.ClientResponse
            status_code = response.status

        if response_json is None:
            try:
                response_json = response.json()
            except requests.JSONDecodeError:
                super().__init__(f"{status_code} {response.reason}")
                return

        errors = response_json.get("errors", [])

        # Use := when support for Python 3.7 is dropped
        if "error" in response_json:
            errors.append(response_json["error"])

        error_text = ""

        for error in errors:
            self.api_errors.append(error)

            if isinstance(error, str):
                self.api_messages.append(error)
                error_text += '\n' + error
                continue

            if "code" in error:
                self.api_codes.append(error["code"])
            if "message" in error:
                self.api_messages.append(error["message"])

            if "code" in error and "message" in error:
                error_text += f"\n{error['code']} - {error['message']}"
            elif "message" in error:
                error_text += '\n' + error["message"]

        # Use := when support for Python 3.7 is dropped
        if not error_text and "detail" in response_json:
            self.api_messages.append(response_json["detail"])
            error_text = '\n' + response_json["detail"]

        super().__init__(
            f"{status_code} {response.reason}{error_text}"
        )


class BadRequest(HTTPException):
    """BadRequest()

    Exception raised for a 400 HTTP status code

    .. version:: 1.0    """
    pass


class Unauthorized(HTTPException):
    """Unauthorized()

    Exception raised for a 401 HTTP status code

    .. version:: 1.0    """
    pass


class Forbidden(HTTPException):
    """Forbidden()

    Exception raised for a 403 HTTP status code

    .. version:: 1.0    """
    pass


class NotFound(HTTPException):
    """NotFound()

    Exception raised for a 404 HTTP status code

    .. version:: 1.0    """
    pass


class TooManyRequests(HTTPException):
    """TooManyRequests()

    Exception raised for a 429 HTTP status code

    .. version:: 1.0
    """
    pass


class KnowAlmostServerError(HTTPException):
    """KnowAlmostServerError()

    Exception raised for a 5xx HTTP status code

    .. version:: 1.0
    """
    pass
